"""
Models module
"""

import os
import os.path

class Models(object):
    """
    Common methods for generating data paths.
    """

    @staticmethod
    def basePath(create=False):
        """
        Base data path - ~/.codequestion

        Args:
            create: if directory should be created

        Returns:
            path
        """

        # Get model cache path
        path = os.path.join(os.path.expanduser("~"), ".codequestion")

        # Create directory if required
        if create:
            os.makedirs(path, exist_ok=True)

        return path

    @staticmethod
    def modelPath(name, create=False):
        """
        Model path for name

        Args:
            name: model name
            create: if directory should be created

        Returns:
            path
        """

        path = os.path.join(Models.basePath(), "models", name)

        # Create directory if required
        if create:
            os.makedirs(path, exist_ok=True)

        return path

    @staticmethod
    def testPath(source, name):
        """
        Builds path to a test data file.

        Args:
            source: test source name
            name: file name

        Returns:
            path
        """

        return os.path.join(Models.basePath(), "test", source, name)

    @staticmethod
    def vectorPath(name, create=False):
        """
        Vector path for name

        Args:
            name: vectors name
            create: if directory should be created

        Returns:
            path
        """

        path = os.path.join(Models.basePath(), "vectors")

        # Create directory path if required
        if create:
            os.makedirs(path, exist_ok=True)

        # Append file name to path
        return os.path.join(path, name)



"""
Text tokenization methods
"""

import re
import string

class Tokenizer(object):
    """
    Text tokenization methods
    """

    # Use standard python punctuation chars but allow tokens to end in # (to allow c#, f#) and + to allow (c++ g++)
    PUNCTUATION = string.punctuation.replace("#", "").replace("+", "")

    # English Stop Word List (Standard stop words used by Apache Lucene)
    STOP_WORDS = {"a", "an", "and", "are", "as", "at", "be", "but", "by", "for", "if", "in", "into", "is", "it",
                  "no", "not", "of", "on", "or", "such", "that", "the", "their", "then", "there", "these",
                  "they", "this", "to", "was", "will", "with"}

    @staticmethod
    def tokenize(text):
        """
        Tokenizes input text into a list of tokens. Filters tokens that match a specific pattern and removes stop words.

        Args:
            text: input text

        Returns:
            list of tokens
        """

        # Convert to all lowercase, split on whitespace, strip punctuation
        tokens = [token.strip(Tokenizer.PUNCTUATION) for token in text.lower().split()]

        # Filter tokens that are numbers or a valid string at least 2 characters long. Remove stop words.
        # Assume tokens already are uncased (all lowercase)
        return [token for token in tokens if (re.match(r"^[#*+\-.0-9:@_a-z]{2,}$", token) or token.isdigit()) and token not in Tokenizer.STOP_WORDS]



"""
Indexing module
"""

import os.path
import sqlite3

from txtai.embeddings import Embeddings

from .models import Models
from .tokenizer import Tokenizer

class Index(object):
    """
    Methods to build a new sentence embeddings index.
    """

    @staticmethod
    def stream(dbfile):
        """
        Streams questions from a questions.db file. This method is a generator and will yield a row at time.

        Args:
            dbfile: input SQLite file
        """

        # Connection to database file
        db = sqlite3.connect(dbfile)
        cur = db.cursor()

        cur.execute("SELECT Id, Question, Source, Tags FROM questions")

        count = 0
        for question in cur:
            # Tokenize question, source and tags
            tokens = Tokenizer.tokenize(question[1] + " " + question[2] + " " + question[3])

            document = (question[0], tokens, question[3])

            count += 1
            if count % 1000 == 0:
                print("Streamed %d documents" % (count), end="\r")

            # Skip documents with no tokens parsed
            if tokens:
                yield document

        print("Iterated over %d total rows" % (count))

        # Free database resources
        db.close()

    @staticmethod
    def embeddings(dbfile):
        """
        Builds a sentence embeddings index.

        Args:
            dbfile: input SQLite file

        Returns:
            embeddings index
        """

        embeddings = Embeddings({"path": Models.vectorPath("stackexchange-300d.magnitude"),
                                 "storevectors": True,
                                 "scoring": "bm25",
                                 "pca": 3,
                                 "quantize": True})

        # Build scoring index if scoring method provided
        if embeddings.config.get("scoring"):
            embeddings.score(Index.stream(dbfile))

        # Build embeddings index
        embeddings.index(Index.stream(dbfile))

        return embeddings

    @staticmethod
    def run():
        """
        Executes an index run.
        """

        path = Models.modelPath("stackexchange")
        dbfile = os.path.join(path, "questions.db")

        print("Building new model")
        embeddings = Index.embeddings(dbfile)
        embeddings.save(path)

if __name__ == "__main__":
    Index.run()



"""
Download module
"""

import os.path
import tempfile
import zipfile

from urllib.request import urlopen

from tqdm import tqdm

from .models import Models

class Download(object):
    """
    Downloads a pre-trained model.
    """

    @staticmethod
    def download(url, dest):
        """
        Downloads a remote file from url and stores at dest.

        Args:
            url: remote url
            dest: destination file path
        """

        with urlopen(url) as response:
            buffer = 16 * 1024
            size = int(response.info()["Content-Length"])

            with tqdm(total=size, unit="B", unit_scale=True, unit_divisor=1024) as pbar:
                with open(dest, "wb") as f:
                    while True:
                        chunk = response.read(buffer)
                        if not chunk:
                            break

                        f.write(chunk)
                        pbar.update(len(chunk))

    @staticmethod
    def run(url):
        """
        Downloads a pre-trained model from url into the local model cache directory.

        Args:
            url: url model path
        """

        # Get base models path
        path = Models.basePath(True)
        dest = os.path.join(tempfile.gettempdir(), os.path.basename(url))

        print("Downloading model from %s to %s" % (url, dest))

        # Download file
        Download.download(url, dest)

        print("Decompressing model to %s" % path)

        # Ensure file was downloaded successfully
        if os.path.exists(dest):
            with zipfile.ZipFile(dest, "r") as z:
                z.extractall(path)

        print("Download complete")

if __name__ == "__main__":
    Download.run("https://github.com/neuml/codequestion/releases/download/v1.1.0/cqmodel.zip")



"""
Evaluation methods to measure model performance
"""

import argparse
import csv
import os
import os.path
import re
import sqlite3

from scipy.stats import pearsonr, spearmanr

from txtai.embeddings import Embeddings

from .models import Models
from .tokenizer import Tokenizer

class Command(object):
    """
    Command line parser
    """

    @staticmethod
    def parse():
        """
        Parses command line arguments.
        """

        parser = argparse.ArgumentParser(description="Evaluate")
        parser.add_argument("-s", "--source", required=True, help="data source", metavar="SOURCE")
        parser.add_argument("-m", "--method", help="run method", metavar="METHOD")

        return parser.parse_args()

class StackExchange(object):
    """
    StackExchange query-answer dataset.
    """

    @staticmethod
    def load():
        """
        Loads a questions database and pre-trained embeddings model

        Returns:
            (db, embeddings)
        """

        print("Loading model")

        path = Models.modelPath("stackexchange")
        dbfile = os.path.join(path, "questions.db")

        # Connect to database file
        db = sqlite3.connect(dbfile)

        # Loading embeddings model
        embeddings = Embeddings()
        embeddings.load(path)

        return db, embeddings

    @staticmethod
    def run(args):
        """
        Evaluates a pre-trained model against the StackExchange query-answer dataset.

        Args:
            args: command line arguments
        """

        # Load model
        db, embeddings = StackExchange.load()
        cur = db.cursor()

        # Statistics
        mrr = []

        # Run test data
        with open(Models.testPath("stackexchange", "query.txt")) as rows:
            for row in rows:
                query, sourceid, source, _ = row.split("|", 3)
                print(query, sourceid, source)

                # Lookup internal id for source id/source
                cur.execute("SELECT Id FROM questions WHERE SourceId = ? and Source=?", [sourceid, source])
                uid = cur.fetchone()
                if uid:
                    uid = uid[0]

                    # Get top 10 results
                    if args.method in ("bm25", "tfidf"):
                        sql = "SELECT Id from search WHERE search=? %s LIMIT 10" % ("ORDER BY BM25(search)" if args.method == "bm25" else "")
                        cur.execute(sql, [re.sub(r"[^\w\s]", "", query)])
                        uids = [row[0] for row in cur.fetchall()]
                    else:
                        # Tokenize query and run search against embeddings model
                        query = Tokenizer.tokenize(query)
                        uids = [uid for uid, _ in embeddings.search(query, 10)]

                    # Calculate stats
                    calc = 1 / (1 + uids.index(uid)) if uid in uids else 0.0
                    print(calc)
                    mrr.append(calc)

        mrr = sum(mrr) / len(mrr)
        print("Mean Reciprocal Rank = ", mrr)

        db.close()

class STS(object):
    """
    STS Benchmark Dataset
    General text similarity

    http://ixa2.si.ehu.es/stswiki/index.php/STSbenchmark
    """

    @staticmethod
    def read(path):
        """
        Reads a STS data file.

        Args:
            path: full path to file

        Returns:
            rows
        """

        data = csv.reader(open(path), delimiter="\t", quoting=csv.QUOTE_NONE)

        rows = []

        # Column Index-Name: 4-score, 5-string 1, 6-string 2
        for x, row in enumerate(data):
            # Normalize score from 0-5 to 0-1. 1 being most similar.
            score = float(row[4]) / 5.0

            # Store row as id (1 indexed), normalized score, string 1, string 2
            rows.append((x + 1, score, row[5], row[6]))

        return rows

    @staticmethod
    def train(vector, score):
        """
        Trains an Embeddings model on STS dev + train data.

        Args:
            vector: word vector model path
            score: scoring method (bm25, sif, tfidf or None for averaging)

        Returns:
            trained Embeddings model
        """

        print("Building model")
        embeddings = Embeddings({"path": Models.vectorPath(vector),
                                 "scoring": score,
                                 "pca": 3})

        rows1 = STS.read(Models.testPath("stsbenchmark", "sts-dev.csv"))
        rows2 = STS.read(Models.testPath("stsbenchmark", "sts-train.csv"))

        rows = rows1 + rows2

        documents = []
        for row in rows:
            tokens = Tokenizer.tokenize(row[2] + " " + row[3])

            if tokens:
                documents.append((row[0], tokens, None))
            else:
                print("Skipping all stop word string: ", row)

        # Build scoring index if scoring method provided
        if embeddings.config.get("scoring"):
            embeddings.score(documents)

        # Build embeddings index
        embeddings.index(documents)

        return embeddings

    @staticmethod
    def test(args, embeddings):
        """
        Tests input Embeddings model against STS benchmark data

        Args:
            args: command line arguments
            embeddings: Embeddings model
        """

        print("Testing model")

        # Test file path
        path = Models.testPath("stsbenchmark", "sts-%s.csv" % ("dev" if args.method == "dev" else "test"))

        # Read test data
        rows = STS.read(path)

        # Calculated scores and ground truth labels
        scores = []
        labels = []

        for row in rows:
            tokens1 = Tokenizer.tokenize(row[2])
            tokens2 = Tokenizer.tokenize(row[3])

            if tokens1 and tokens2:
                score = embeddings.similarity(tokens1, [tokens2])[0][1]
                scores.append(score)

                # Ground truth score normalized between 0 - 1
                labels.append(row[1])
            else:
                print("Skipping all stop word string: ", row)

        print("Pearson score =", pearsonr(scores, labels))
        print("Spearman score =", spearmanr(scores, labels))

    @staticmethod
    def run(args):
        """
        Runs multiple combinations of vector and score combinations.

        Args:
            args: command line arguments
        """

        vectors = ["stackexchange-300d.magnitude"]
        scoring = ["bm25"]

        for vector in vectors:
            for score in scoring:
                print("%s - %s" % (vector, score))
                embeddings = STS.train(vector, score)
                STS.test(args, embeddings)

if __name__ == "__main__":
    # Command line parser
    ARGS = Command.parse()

    if ARGS.source.lower() == "sts":
        STS.run(ARGS)
    else:
        StackExchange.run(ARGS)



"""
codequestion query shell module.
"""

from cmd import Cmd

from .query import Query

class Shell(Cmd):
    """
    codequestion query shell.
    """

    intro = "codequestion query shell"
    prompt = "(cqq) "
    embeddings = None
    db = None
    lang = None

    def preloop(self):
        # Load embeddings and questions.db
        self.embeddings, self.db = Query.load()

    def postloop(self):
        if self.db:
            self.db.close()

    def default(self, line):
        Query.query(self.embeddings, self.db, "%s %s" % (self.lang, line) if self.lang else line)

    def do_lang(self, lang):
        """
        Sets the default programming language for this session. All queries will have the language
        prepended to the query.

        Args:
            lang: language
        """

        self.lang = lang
        print("Set language to %s" % lang)

def main():
    """
    Shell execution loop.
    """

    Shell().cmdloop()

if __name__ == "__main__":
    main()



"""
Vectors module
"""

import os
import os.path
import sqlite3
import tempfile

from txtai.vectors import WordVectors

from .models import Models
from .tokenizer import Tokenizer

class RowIterator(object):
    """
    Iterates over rows in a database query. Allows for multiple iterations.
    """

    def __init__(self, dbfile):
        """
        Initializes RowIterator.

        Args:
            dbfile: path to SQLite file
        """

        # Store database file
        self.dbfile = dbfile

        self.rows = self.stream(self.dbfile)

    def __iter__(self):
        """
        Creates a database query generator.

        Returns:
            generator
        """

        # reset the generator
        self.rows = self.stream(self.dbfile)
        return self

    def __next__(self):
        """
        Gets the next result in the current generator.

        Returns:
            tokens
        """

        result = next(self.rows)
        if result is None:
            raise StopIteration
        else:
            return result

    def stream(self, dbfile):
        """
        Connects to SQLite file at dbfile and yields parsed tokens for each row.

        Args:
            dbfile:
        """

        # Connection to database file
        db = sqlite3.connect(dbfile)
        cur = db.cursor()

        cur.execute("SELECT Question, Source, Tags FROM questions")

        count = 0
        for question in cur:
            # Tokenize question, source and tags
            tokens = Tokenizer.tokenize(question[0] + " " + question[1] + " " + question[2])

            count += 1
            if count % 1000 == 0:
                print("Streamed %d documents" % (count), end="\r")

            # Skip documents with no tokens parsed
            if tokens:
                yield tokens

        print("Iterated over %d total rows" % (count))

        # Free database resources
        db.close()

class Vectors(object):
    """
    Methods to build a FastText model.
    """

    @staticmethod
    def tokens(dbfile):
        """
        Iterates over each row in dbfile and writes parsed tokens to a temporary file for processing.

        Args:
            dbfile: SQLite file to read

        Returns:
            path to output file
        """

        tokens = None

        # Stream tokens to temp working file
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as output:
            # Save file path
            tokens = output.name

            for row in RowIterator(dbfile):
                output.write(" ".join(row) + "\n")

        return tokens

    @staticmethod
    def run(dbfile, size, mincount):
        """
        Converts dbfile into a fastText model using pymagnitude's SQLite output format.

        Args:
            dbfile: input SQLite file
            size: dimensions for fastText model
            mincount: minimum number of times a token must appear in input
        """

        # Stream tokens to temporary file
        tokens = Vectors.tokens(dbfile)

        # Output file path
        path = Models.vectorPath("stackexchange-%dd" % size, True)

        # Build word vectors model
        WordVectors.build(tokens, size, mincount, path)

        # Remove temporary tokens file
        os.remove(tokens)

if __name__ == "__main__":
    # Resolve questions.db path and run
    Vectors.run(os.path.join(Models.modelPath("stackexchange"), "questions.db"), 300, 3)



"""
Query module
"""

import argparse
import os
import os.path
import sqlite3

import html2text
import mdv

from txtai.embeddings import Embeddings

from .models import Models
from .tokenizer import Tokenizer

class Query(object):
    """
    Methods to query an embeddings index.
    """

    @staticmethod
    def escape(text):
        """
        Escapes text to work around issues with mdv double escaping characters.

        Args:
            text: input text

        Returns:
            escaped text
        """

        text = text.replace("<", "¿")
        text = text.replace(">", "Ñ")
        text = text.replace("&", "ž")

        return text

    @staticmethod
    def unescape(text):
        """
        Un-escapes text to work around issues with mdv double escaping characters.

        Args:
            text: input text

        Returns:
            unescaped text
        """

        text = text.replace("¿", "<")
        text = text.replace("Ñ", ">")
        text = text.replace("ž", "&")

        return text

    @staticmethod
    def render(text, html=True, tab_length=0):
        """
        Renders input text to formatted text ready to send to the terminal.

        Args:
            text: input html text

        Returns:
            text formatted for print to terminal
        """

        if html:
            # Convert HTML
            parser = html2text.HTML2Text()
            parser.body_width = 0
            text = parser.handle(text)

            text = Query.escape(text)

        text = mdv.main(text, theme="592.2129", c_theme="953.3567", cols=180, tab_length=tab_length)

        if html:
            text = Query.unescape(text)

        return text.strip()

    @staticmethod
    def match(query, tags, question):
        """
        Analyzses a query match to determine if it will be accepted as a match.

        Args:
            query: query tokens
            tags: all tags
            question: matching question

        Return:
            True if query is accepted as a match, False otherwise
        """

        # Flatten question object to string for searching
        question = Tokenizer.tokenize(question[0] + " " + question[1] + " " + question[2])

        match = True
        for token in query:
            # Look for tag tokens and require all those tag tokens to be in the matching question
            if token in tags and token not in question:
                match = False
                break

        return match

    @staticmethod
    def load():
        """
        Loads an embeddings model and questions.db database.

        Returns:
            (embeddings, db handle)
        """

        path = Models.modelPath("stackexchange")
        dbfile = os.path.join(path, "questions.db")

        if os.path.isfile(os.path.join(path, "config")):
            print("Loading model from %s" % path)
            embeddings = Embeddings()
            embeddings.load(path)
        else:
            print("ERROR: loading model: ensure model is installed")
            print("ERROR: Pre-trained model can be installed by running python -m codequestion.download")
            raise FileNotFoundError("Unable to load codequestion model from %s" % path)

        # Connect to database file
        db = sqlite3.connect(dbfile)

        return (embeddings, db)

    @staticmethod
    def query(embeddings, db, query):
        """
        Executes a query against the embeddings model.

        Args:
            embeddings: embeddings model
            db: open SQLite database
            query: query string
        """

        cur = db.cursor()

        query = Tokenizer.tokenize(query)
        print("Query: ", query)

        accepted = 0
        for uid, score in embeddings.search(query, 10):
            cur.execute("SELECT QuestionUser, Question, Tags, AnswerUser, Answer, Reference FROM questions WHERE id = ?", [uid])

            question = cur.fetchone()
            match = Query.match(query, embeddings.scoring.tags, question)
            if match or score >= 0.9:
                print(Query.render("#Question (by %s): %s [%.4f]" % (question[0], question[1], score), html=False))
                print("Tags: %s" % question[2])
                print("Answer (by %s):\n%s" % (question[3], Query.render(question[4], 2)))
                print("\nReference: %s" % question[5])

                accepted += 1
                if match or accepted >= 2:
                    break

                print()

    @staticmethod
    def close(db):
        """
        Closes a SQLite database database.

        Args:
            db: open database
        """

        # Free database resources
        db.close()

    @staticmethod
    def args():
        """
        Parses command line arguments.

        Returns:
            command line arguments
        """

        parser = argparse.ArgumentParser(description="codequestion query")

        parser.add_argument("-l", "--lang", help="Sets default language")
        parser.add_argument("-q", "--query", required=True, help="Query to execute")

        return parser.parse_args()

    @staticmethod
    def run():
        """
        Executes a query against a codequestion index.
        """

        # Parse command line arguments
        args = Query.args()

        # Build query
        query = "%s %s" % (args.lang, args.query) if args.lang else args.query

        # Load model
        embeddings, db = Query.load()

        # Query the database
        Query.query(embeddings, db, query)

        # Free resources
        Query.close(db)

if __name__ == "__main__":
    Query.run()
